import React, { Component, MouseEvent, ChangeEvent } from 'react';


/** Represents an item in a Todo list. */
type TodoItem = {
  /** Name of the todo item. */
  name: string;

  /** Indicates if the todo item has been completed. */
  completed: boolean;
};

type TodoState = {
  /** Current TodoItems */
  items: TodoItem[];

  /** Mirrors text in the new todo item text box. */
  newItemName: string;
}

// Enable debug logging of events in this component.
const DEBUG: boolean = false;

/** Displays the todo UI. Allows users to add/complete items on a todo list. */
export class TodoApp extends Component<{}, TodoState> {
  constructor(props: {}) {
    super(props);

    this.state = {items: [], newItemName: ""};
  }

  render = (): JSX.Element => {
    if (DEBUG) console.debug("rendering");  // Tip: A great place to log the state!

    return (
      <div>
        <h2>To-Do List</h2>
        {this.renderItems()}
        <p>Check the item to mark it completed.</p>
        <label htmlFor="name">New item: </label>
        <input id="name" type="text" value={this.state.newItemName}
            onChange={this.doNewItemNameChange} />
        <button type="button" onClick={this.doAddClick}>Add</button>
      </div>);
  };

  // Returns a div for each item in the todo list, with each div containing a
  // label with the name of the item and a checkbox for marking it completed.
  renderItems = (): JSX.Element[] => {
    const items: JSX.Element[] = [];
    // Inv: items contains a div for each item of this.state.items[0 .. i-1]
    for (const item of this.state.items) {
      const i = this.state.items.findIndex((i) => i === item);
      const id = "check_" + i;
      if (item.completed) {
        items.push(
          <div key={i}>
            <input type="checkbox" id={id} checked={true} readOnly={true} />
            <label className="completed" htmlFor={id}>{item.name}</label>
          </div>);
      } else {
        items.push(
          <div key={i}>
            <input type="checkbox" id={id} checked={false}
                onChange={evt => this.doItemClick(evt, i)} />
            <label htmlFor={id}>{item.name}</label>
          </div>);
      }
    }
    return items;
  };

  // Called when the user checks the box next to an uncompleted item. index is
  // the index of that item in the list.
  doItemClick = (_: ChangeEvent<HTMLInputElement>, index: number): void => {
    const item = this.state.items[index];
    if (DEBUG) console.log(`marking item ${item.name} as completed`);

    // Note: we cannot mutate the list. We have to create a new one.
    const newItems = this.state.items.slice(0, index)
        .concat([{name: item.name, completed: true}])
        .concat(this.state.items.slice(index + 1));
    this.setState({items: newItems});

    setTimeout(() => this.doItemTimeout(index), 5000);
  };

  // Called after an item has been marked as completed for 5 seconds.
  doItemTimeout = (index: number): void => {
    const item = this.state.items[index];
    if (DEBUG) console.log(`removing item ${item.name}`);

    // Note: we cannot mutate the list. We have to create a new one.
    const items = this.state.items.slice(0, index)
        .concat(this.state.items.slice(index + 1));
    this.setState({items: items});
  };

  // Called when the user clicks on the button to add the new item.
  doAddClick = (_: MouseEvent<HTMLButtonElement>): void => {
    // Ignore the request if the user hasn't entered a name.
    // Note: Would be better to disable the button in this case.
    const name = this.state.newItemName.trim();
    if (name.length == 0)
      return;

    if (DEBUG) console.log(`adding new item ${name}`);

    const newItems = this.state.items.concat([{name: name, completed: false}]);
    this.setState({items: newItems, newItemName: ""});
  };

  // Called each time the text in the new item name field is changed.
  doNewItemNameChange = (evt: ChangeEvent<HTMLInputElement>): void => {
    if (DEBUG) console.log(`changing new name to ${evt.target.value}`);
    this.setState({newItemName: evt.target.value});
  };
}
